@@ -15,7 +15,7 @@
   | Author: Sara Golemon <pollita@php.net>                               |
   +----------------------------------------------------------------------+
 
-  $Id: ssh2.c 317115 2011-09-21 17:40:23Z bjori $ 
+  $Id: ssh2.c 327906 2012-10-06 13:52:51Z langemeijer $ 
 */
 
 #ifdef HAVE_CONFIG_H
@@ -43,13 +43,9 @@
 
 /* True global resources - no need for thread safety here */
 int le_ssh2_session;
-#ifdef PHP_SSH2_REMOTE_FORWARDING
 int le_ssh2_listener;
-#endif
 int le_ssh2_sftp;
-#ifdef PHP_SSH2_PUBLICKEY_SUBSYSTEM
 int le_ssh2_pkey_subsys;
-#endif
 
 ZEND_BEGIN_ARG_INFO(php_ssh2_first_arg_force_ref, 0)
     ZEND_ARG_PASS_INFO(1)
@@ -240,6 +236,8 @@ LIBSSH2_DISCONNECT_FUNC(php_ssh2_disconnect_cb)
 }
 /* }}} */
 
+
+
 /* *****************
    * Userspace API *
    ***************** */
@@ -455,7 +453,7 @@ PHP_FUNCTION(ssh2_connect)
 	int host_len;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "s|la!a!", &host, &host_len, &port, &methods, &callbacks) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	session = php_ssh2_session_connect(host, port, methods, callbacks TSRMLS_CC);
@@ -478,14 +476,11 @@ PHP_FUNCTION(ssh2_methods_negotiated)
 	char *kex, *hostkey, *crypt_cs, *crypt_sc, *mac_cs, *mac_sc, *comp_cs, *comp_sc, *lang_cs, *lang_sc;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zsession) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
 
-#if defined(LIBSSH2_APINO) && LIBSSH2_APINO < 200412301450
-	libssh2_session_methods(session, &kex, &hostkey, &crypt_cs, &crypt_sc, &mac_cs, &mac_sc, &comp_cs, &comp_sc, &lang_cs, &lang_sc);
-#else
 	kex = (char*)libssh2_session_methods(session, LIBSSH2_METHOD_KEX);
 	hostkey = (char*)libssh2_session_methods(session, LIBSSH2_METHOD_HOSTKEY);
 	crypt_cs = (char*)libssh2_session_methods(session, LIBSSH2_METHOD_CRYPT_CS);
@@ -496,7 +491,6 @@ PHP_FUNCTION(ssh2_methods_negotiated)
 	comp_sc = (char*)libssh2_session_methods(session, LIBSSH2_METHOD_COMP_SC);
 	lang_cs = (char*)libssh2_session_methods(session, LIBSSH2_METHOD_LANG_CS);
 	lang_sc = (char*)libssh2_session_methods(session, LIBSSH2_METHOD_LANG_SC);
-#endif
 
 	array_init(return_value);
 	add_assoc_string(return_value, "kex", kex, 1);
@@ -533,7 +527,7 @@ PHP_FUNCTION(ssh2_fingerprint)
 	int i, fingerprint_len;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r|l", &zsession, &flags) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 	fingerprint_len = (flags & PHP_SSH2_FINGERPRINT_SHA1) ? SHA_DIGEST_LENGTH : MD5_DIGEST_LENGTH;
 
@@ -579,7 +573,7 @@ PHP_FUNCTION(ssh2_auth_none)
 	int username_len;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zsession, &username, &username_len) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
@@ -603,6 +597,27 @@ PHP_FUNCTION(ssh2_auth_none)
 }
 /* }}} */
 
+char *password_for_kbd_callback;
+
+static void kbd_callback(const char *name, int name_len,
+							const char *instruction, int instruction_len,
+							int num_prompts,
+							const LIBSSH2_USERAUTH_KBDINT_PROMPT *prompts,
+							LIBSSH2_USERAUTH_KBDINT_RESPONSE *responses,
+							void **abstract)
+{
+	(void)name;
+	(void)name_len;
+	(void)instruction;
+	(void)instruction_len;
+	if (num_prompts == 1) {
+		responses[0].text = estrdup(password_for_kbd_callback);
+		responses[0].length = strlen(password_for_kbd_callback);
+	}
+	(void)prompts;
+	(void)abstract;
+}
+
 /* {{{ proto bool ssh2_auth_password(resource session, string username, string password)
  * Authenticate over SSH using a plain password
  */
@@ -612,12 +627,21 @@ PHP_FUNCTION(ssh2_auth_password)
 	zval *zsession;
 	char *username, *password;
 	int username_len, password_len;
+	char *userauthlist;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss", &zsession, &username, &username_len, &password, &password_len) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_NONAUTHENTICATED_SESSION(session, zsession);
+
+	userauthlist = libssh2_userauth_list(session, username, username_len);
+	password_for_kbd_callback = password;
+	if (strstr(userauthlist, "keyboard-interactive") != NULL) {
+		if (libssh2_userauth_keyboard_interactive(session, username, &kbd_callback) == 0) {
+			RETURN_TRUE;
+		}
+	}
 
 	/* TODO: Support password change callback */
 	if (libssh2_userauth_password_ex(session, username, username_len, password, password_len, NULL)) {
@@ -638,19 +662,39 @@ PHP_FUNCTION(ssh2_auth_pubkey_file)
 	zval *zsession;
 	char *username, *pubkey, *privkey, *passphrase = NULL;
 	int username_len, pubkey_len, privkey_len, passphrase_len;
+	char *newpath;
+	struct passwd *pws;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsss|s", &zsession,	&username, &username_len,
 																				&pubkey, &pubkey_len,
 																				&privkey, &privkey_len,
 																				&passphrase, &passphrase_len) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	if (SSH2_OPENBASEDIR_CHECKPATH(pubkey) || SSH2_OPENBASEDIR_CHECKPATH(privkey)) {
 		RETURN_FALSE;
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_NONAUTHENTICATED_SESSION(session, zsession);
+
+	// Explode '~/paths' stopgap fix because libssh2 does not accept tilde for homedir
+	// This should be ifdef'ed when a fix is available to support older libssh2 versions
+	pws = getpwuid(geteuid());
+	if (pubkey_len >= 2 && *pubkey == '~' && *(pubkey+1) == '/') {
+		newpath = emalloc(strlen(pws->pw_dir) + strlen(pubkey));
+		strcpy(newpath, pws->pw_dir);
+		strcat(newpath, pubkey+1);
+		efree(pubkey);
+		pubkey = newpath;
+	}
+	if (privkey_len >= 2 && *privkey == '~' && *(privkey+1) == '/') {
+		newpath = emalloc(strlen(pws->pw_dir) + strlen(privkey));
+		strcpy(newpath, pws->pw_dir);
+		strcat(newpath, privkey+1);
+		efree(privkey);
+		privkey = newpath;
+	}
 
 	/* TODO: Support passphrase callback */
 	if (libssh2_userauth_publickey_fromfile_ex(session, username, username_len, pubkey, privkey, passphrase)) {
@@ -665,7 +709,6 @@ PHP_FUNCTION(ssh2_auth_pubkey_file)
 }
 /* }}} */
 
-#ifdef PHP_SSH2_HOSTBASED_AUTH
 /* {{{ proto bool ssh2_auth_hostbased_file(resource session, string username, string local_hostname, string pubkeyfile, string privkeyfile[, string passphrase[, string local_username]])
  * Authenticate using a hostkey
  */
@@ -682,14 +725,14 @@ PHP_FUNCTION(ssh2_auth_hostbased_file)
 																					&privkey, &privkey_len,
 																					&passphrase, &passphrase_len,
 																					&local_username, &local_username_len) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	if (SSH2_OPENBASEDIR_CHECKPATH(pubkey) || SSH2_OPENBASEDIR_CHECKPATH(privkey)) {
 		RETURN_FALSE;
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_NONAUTHENTICATED_SESSION(session, zsession);
 
 	if (!local_username) {
 		local_username = username;
@@ -705,9 +748,7 @@ PHP_FUNCTION(ssh2_auth_hostbased_file)
 	RETURN_TRUE;
 }
 /* }}} */
-#endif /* PHP_SSH2_HOSTBASED_AUTH */
 
-#ifdef PHP_SSH2_REMOTE_FORWARDING
 /* {{{ proto resource ssh2_forward_listen(resource session, int port[, string host[, long max_connections]])
  * Bind a port on the remote server and listen for connections
  */
@@ -723,10 +764,10 @@ PHP_FUNCTION(ssh2_forward_listen)
 	long max_connections = PHP_SSH2_LISTEN_MAX_QUEUED;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl|sl", &zsession, &port, &host, &host_len, &max_connections) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_AUTHENTICATED_SESSION(session, zsession);
 
 	listener = libssh2_channel_forward_listen_ex(session, host, port, NULL, max_connections);	
 
@@ -757,7 +798,7 @@ PHP_FUNCTION(ssh2_forward_accept)
 	php_stream *stream;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zlistener) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	ZEND_FETCH_RESOURCE(data, php_ssh2_listener_data*, &zlistener, -1, PHP_SSH2_LISTENER_RES_NAME, le_ssh2_listener);
@@ -787,9 +828,7 @@ PHP_FUNCTION(ssh2_forward_accept)
 	php_stream_to_zval(stream, return_value);
 }
 /* }}} */
-#endif /* PHP_SSH2_REMOTE_FORWARDING */
 
-#ifdef PHP_SSH2_POLL
 /* {{{ proto int ssh2_poll(array &polldes[, int timeout])
  * Poll the channels/listeners/streams for events
  * Returns number of descriptors which returned non-zero revents
@@ -814,7 +853,7 @@ PHP_FUNCTION(ssh2_poll)
 	zval ***pollmap;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "a|l", &zdesc, &timeout) == FAILURE) {
-		RETURN_NULL();
+		return;
 	}
 
 	numfds = zend_hash_num_elements(Z_ARRVAL_P(zdesc));
@@ -893,9 +932,7 @@ PHP_FUNCTION(ssh2_poll)
 	RETURN_LONG(fds_ready);
 }
 /* }}} */
-#endif /* PHP_SSH2_POLL */
 
-#ifdef PHP_SSH2_PUBLICKEY_SUBSYSTEM
 /* ***********************
    * Publickey Subsystem *
    *********************** */
@@ -910,10 +947,10 @@ PHP_FUNCTION(ssh2_publickey_init)
 	php_ssh2_pkey_subsys_data *data;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zsession) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_AUTHENTICATED_SESSION(session, zsession);
 
 	pkey = libssh2_publickey_init(session);
 
@@ -949,7 +986,7 @@ PHP_FUNCTION(ssh2_publickey_add)
 	zend_bool overwrite = 0;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss|ba", &zpkey_data, &algo, &algo_len, &blob, &blob_len, &overwrite, &zattrs) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	ZEND_FETCH_RESOURCE(data, php_ssh2_pkey_subsys_data*, &zpkey_data, -1, PHP_SSH2_PKEY_SUBSYS_RES_NAME, le_ssh2_pkey_subsys);
@@ -1040,7 +1077,7 @@ PHP_FUNCTION(ssh2_publickey_remove)
 	int algo_len, blob_len;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss", &zpkey_data, &algo, &algo_len, &blob, &blob_len) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	ZEND_FETCH_RESOURCE(data, php_ssh2_pkey_subsys_data*, &zpkey_data, -1, PHP_SSH2_PKEY_SUBSYS_RES_NAME, le_ssh2_pkey_subsys);
@@ -1064,7 +1101,7 @@ PHP_FUNCTION(ssh2_publickey_list)
 	libssh2_publickey_list *keys;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zpkey_data) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	ZEND_FETCH_RESOURCE(data, php_ssh2_pkey_subsys_data*, &zpkey_data, -1, PHP_SSH2_PKEY_SUBSYS_RES_NAME, le_ssh2_pkey_subsys);
@@ -1102,7 +1139,87 @@ PHP_FUNCTION(ssh2_publickey_list)
 	libssh2_publickey_list_free(data->pkey, keys);
 }
 /* }}} */
-#endif /* PHP_SSH2_PUBLICKEY_SUBSYSTEM */
+
+/* {{{ proto array ssh2_auth_agent(resource session, string username)
+Authenticate using the ssh agent */
+PHP_FUNCTION(ssh2_auth_agent)
+{
+#ifdef PHP_SSH2_AGENT_AUTH
+	zval *zsession;
+	char *username;
+	int username_len;
+
+	LIBSSH2_SESSION *session;
+	char *userauthlist;
+	LIBSSH2_AGENT *agent = NULL;
+	int rc;
+	struct libssh2_agent_publickey *identity, *prev_identity = NULL;
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zsession, &username, &username_len) == FAILURE) {
+		return;
+	}
+
+	SSH2_FETCH_NONAUTHENTICATED_SESSION(session, zsession);
+
+	/* check what authentication methods are available */
+	userauthlist = libssh2_userauth_list(session, username, username_len);
+
+	if (strstr(userauthlist, "publickey") == NULL) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "\"publickey\" authentication is not supported");
+		RETURN_FALSE;
+	}
+
+	/* Connect to the ssh-agent */
+	agent = libssh2_agent_init(session);
+
+	if (!agent) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failure initializing ssh-agent support");
+		RETURN_FALSE;
+	}
+
+	if (libssh2_agent_connect(agent)) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failure connecting to ssh-agent");
+		libssh2_agent_free(agent);
+		RETURN_FALSE;
+	}
+
+	if (libssh2_agent_list_identities(agent)) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failure requesting identities to ssh-agent");
+		libssh2_agent_disconnect(agent);
+		libssh2_agent_free(agent);
+		RETURN_FALSE;
+	}
+
+	while (1) {
+		rc = libssh2_agent_get_identity(agent, &identity, prev_identity);
+
+		if (rc == 1) {
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Couldn't continue authentication");
+			libssh2_agent_disconnect(agent);
+			libssh2_agent_free(agent);
+			RETURN_FALSE;
+		}
+
+		if (rc < 0) {
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failure obtaining identity from ssh-agent support");
+			libssh2_agent_disconnect(agent);
+			libssh2_agent_free(agent);
+			RETURN_FALSE;
+		}
+
+		if (!libssh2_agent_userauth(agent, username, identity)) {
+			libssh2_agent_disconnect(agent);
+			libssh2_agent_free(agent);
+			RETURN_TRUE;
+		}
+		prev_identity = identity;
+	}
+#else
+	php_error_docref(NULL TSRMLS_CC, E_WARNING, "Upgrade the libssh2 library (needs 1.2.3 or higher) and reinstall the ssh2 extension for ssh2 agent support");
+	RETURN_FALSE;
+#endif /* PHP_SSH2_AGENT_AUTH */
+}
+/* }}} */
 
 /* ***********************
    * Module Housekeeping *
@@ -1138,7 +1255,6 @@ static void php_ssh2_session_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC)
 	libssh2_session_free(session);
 }
 
-#ifdef PHP_SSH2_REMOTE_FORWARDING
 static void php_ssh2_listener_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC)
 {
 	php_ssh2_listener_data *data = (php_ssh2_listener_data*)rsrc->ptr;
@@ -1148,9 +1264,7 @@ static void php_ssh2_listener_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC)
 	zend_list_delete(data->session_rsrcid);
 	efree(data);
 }
-#endif /* PHP_SSH2_REMOTE_FORWARDING */
 
-#ifdef PHP_SSH2_PUBLICKEY_SUBSYSTEM
 static void php_ssh2_pkey_subsys_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC)
 {
 	php_ssh2_pkey_subsys_data *data = (php_ssh2_pkey_subsys_data*)rsrc->ptr;
@@ -1160,20 +1274,15 @@ static void php_ssh2_pkey_subsys_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC)
 	zend_list_delete(data->session_rsrcid);
 	efree(data);
 }
-#endif /* PHP_SSH2_PUBLICKEY_SUBSYSTEM */
 
 /* {{{ PHP_MINIT_FUNCTION
  */
 PHP_MINIT_FUNCTION(ssh2)
 {
 	le_ssh2_session		= zend_register_list_destructors_ex(php_ssh2_session_dtor, NULL, PHP_SSH2_SESSION_RES_NAME, module_number);
-#ifdef PHP_SSH2_REMOTE_FORWARDING
 	le_ssh2_listener	= zend_register_list_destructors_ex(php_ssh2_listener_dtor, NULL, PHP_SSH2_LISTENER_RES_NAME, module_number);
-#endif /* PHP_SSH2_REMOTE_FORWARDING */
 	le_ssh2_sftp		= zend_register_list_destructors_ex(php_ssh2_sftp_dtor, NULL, PHP_SSH2_SFTP_RES_NAME, module_number);
-#ifdef PHP_SSH2_PUBLICKEY_SUBSYSTEM
 	le_ssh2_pkey_subsys	= zend_register_list_destructors_ex(php_ssh2_pkey_subsys_dtor, NULL, PHP_SSH2_PKEY_SUBSYS_RES_NAME, module_number);
-#endif /* PHP_SSH2_PUBLICKEY_SUBSYSTEM */
 
 	REGISTER_LONG_CONSTANT("SSH2_FINGERPRINT_MD5",		PHP_SSH2_FINGERPRINT_MD5,		CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("SSH2_FINGERPRINT_SHA1",		PHP_SSH2_FINGERPRINT_SHA1,		CONST_CS | CONST_PERSISTENT);
@@ -1191,7 +1300,6 @@ PHP_MINIT_FUNCTION(ssh2)
 	REGISTER_LONG_CONSTANT("SSH2_STREAM_STDIO",			0,								CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("SSH2_STREAM_STDERR",		SSH_EXTENDED_DATA_STDERR,		CONST_CS | CONST_PERSISTENT);
 
-#ifdef PHP_SSH2_POLL
 	/* events/revents */
 	REGISTER_LONG_CONSTANT("SSH2_POLLIN",				LIBSSH2_POLLFD_POLLIN,			CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("SSH2_POLLEXT",				LIBSSH2_POLLFD_POLLEXT,			CONST_CS | CONST_PERSISTENT);
@@ -1201,12 +1309,9 @@ PHP_MINIT_FUNCTION(ssh2)
 	REGISTER_LONG_CONSTANT("SSH2_POLLERR",				LIBSSH2_POLLFD_POLLERR,			CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("SSH2_POLLHUP",				LIBSSH2_POLLFD_POLLHUP,			CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("SSH2_POLLNVAL",				LIBSSH2_POLLFD_POLLNVAL,		CONST_CS | CONST_PERSISTENT);
-#if (defined(LIBSSH2_APINO) && (LIBSSH2_APINO > 200503221619)) || (defined(LIBSSH2_VERSION_NUM) && LIBSSH2_VERSION_NUM >= 0x001000)
 	REGISTER_LONG_CONSTANT("SSH2_POLL_SESSION_CLOSED",	LIBSSH2_POLLFD_SESSION_CLOSED,	CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("SSH2_POLL_CHANNEL_CLOSED",	LIBSSH2_POLLFD_CHANNEL_CLOSED,	CONST_CS | CONST_PERSISTENT);
 	REGISTER_LONG_CONSTANT("SSH2_POLL_LISTENER_CLOSED",	LIBSSH2_POLLFD_LISTENER_CLOSED,	CONST_CS | CONST_PERSISTENT);
-#endif /* >= LIBSSH2-0.10 */
-#endif /* POLL */
 
 	return (php_register_url_stream_wrapper("ssh2.shell", &php_ssh2_stream_wrapper_shell TSRMLS_CC) == SUCCESS &&
 			php_register_url_stream_wrapper("ssh2.exec", &php_ssh2_stream_wrapper_exec TSRMLS_CC) == SUCCESS &&
@@ -1237,18 +1342,6 @@ PHP_MINFO_FUNCTION(ssh2)
 	php_info_print_table_row(2, "extension version", PHP_SSH2_VERSION);
 	php_info_print_table_row(2, "libssh2 version", LIBSSH2_VERSION);
 	php_info_print_table_row(2, "banner", LIBSSH2_SSH_BANNER);
-#ifdef PHP_SSH2_REMOTE_FORWARDING
-	php_info_print_table_row(2, "remote forwarding", "enabled");
-#endif
-#ifdef PHP_SSH2_HOSTBASED_AUTH
-	php_info_print_table_row(2, "hostbased auth", "enabled");
-#endif
-#ifdef PHP_SSH2_POLL
-	php_info_print_table_row(2, "polling support", "enabled");
-#endif
-#ifdef PHP_SSH2_PUBLICKEY_SUBSYSTEM
-	php_info_print_table_row(2, "publickey subsystem", "enabled");
-#endif
 	php_info_print_table_end();
 }
 /* }}} */
@@ -1263,14 +1356,10 @@ zend_function_entry ssh2_functions[] = {
 	PHP_FE(ssh2_auth_none,						NULL)
 	PHP_FE(ssh2_auth_password,					NULL)
 	PHP_FE(ssh2_auth_pubkey_file,				NULL)
-#ifdef PHP_SSH2_HOSTBASED_AUTH
 	PHP_FE(ssh2_auth_hostbased_file,			NULL)
-#endif /* PHP_SSH2_HOSTBASED_AUTH */
 
-#ifdef PHP_SSH2_REMOTE_FORWARDING
 	PHP_FE(ssh2_forward_listen,					NULL)
 	PHP_FE(ssh2_forward_accept,					NULL)
-#endif /* PHP_SSH2_REMOTE_FORWARDING */
 
 	/* Stream Stuff */
 	PHP_FE(ssh2_shell,							NULL)
@@ -1279,9 +1368,7 @@ zend_function_entry ssh2_functions[] = {
 	PHP_FE(ssh2_scp_recv,						NULL)
 	PHP_FE(ssh2_scp_send,						NULL)
 	PHP_FE(ssh2_fetch_stream,					NULL)
-#ifdef PHP_SSH2_POLL
 	PHP_FE(ssh2_poll,							php_ssh2_first_arg_force_ref)
-#endif
 
 	/* SFTP Stuff */
 	PHP_FE(ssh2_sftp,							NULL)
@@ -1291,6 +1378,7 @@ zend_function_entry ssh2_functions[] = {
 	PHP_FE(ssh2_sftp_unlink,					NULL)
 	PHP_FE(ssh2_sftp_mkdir,						NULL)
 	PHP_FE(ssh2_sftp_rmdir,						NULL)
+	PHP_FE(ssh2_sftp_chmod,						NULL)
 	PHP_FE(ssh2_sftp_stat,						NULL)
 	PHP_FE(ssh2_sftp_lstat,						NULL)
 	PHP_FE(ssh2_sftp_symlink,					NULL)
@@ -1298,12 +1386,12 @@ zend_function_entry ssh2_functions[] = {
 	PHP_FE(ssh2_sftp_realpath,					NULL)
 
 	/* Publickey subsystem */
-#ifdef PHP_SSH2_PUBLICKEY_SUBSYSTEM
 	PHP_FE(ssh2_publickey_init,					NULL)
 	PHP_FE(ssh2_publickey_add,					NULL)
 	PHP_FE(ssh2_publickey_remove,				NULL)
 	PHP_FE(ssh2_publickey_list,					NULL)
-#endif
+
+	PHP_FE(ssh2_auth_agent,						NULL)
 
 	{NULL, NULL, NULL}
 };
