@@ -15,7 +15,7 @@
   | Author: Sara Golemon <pollita@php.net>                               |
   +----------------------------------------------------------------------+
 
-  $Id: ssh2_fopen_wrappers.c 306892 2010-12-30 23:59:55Z pajoye $ 
+  $Id: ssh2_fopen_wrappers.c 327861 2012-10-02 20:18:07Z langemeijer $ 
 */
 
 #ifdef HAVE_CONFIG_H
@@ -32,21 +32,79 @@
 static size_t php_ssh2_channel_stream_write(php_stream *stream, const char *buf, size_t count TSRMLS_DC)
 {
 	php_ssh2_channel_data *abstract = (php_ssh2_channel_data*)stream->abstract;
+	size_t writestate;
+	LIBSSH2_SESSION *session;
 
 	libssh2_channel_set_blocking(abstract->channel, abstract->is_blocking);
-	return libssh2_channel_write_ex(abstract->channel, abstract->streamid, buf, count);
+	session = (LIBSSH2_SESSION *)zend_fetch_resource(NULL TSRMLS_CC, abstract->session_rsrc, PHP_SSH2_SESSION_RES_NAME, NULL, 1, le_ssh2_session);
+
+#ifdef PHP_SSH2_SESSION_TIMEOUT
+	if (abstract->is_blocking) {
+		libssh2_session_set_timeout(session, abstract->timeout);
+	}
+#endif
+
+	writestate = libssh2_channel_write_ex(abstract->channel, abstract->streamid, buf, count);
+
+#ifdef PHP_SSH2_SESSION_TIMEOUT
+	if (abstract->is_blocking) {
+		libssh2_session_set_timeout(session, 0);
+	}
+#endif
+	if (writestate == LIBSSH2_ERROR_EAGAIN) {
+		writestate = 0;
+	}
+
+	if (writestate < 0) {
+		char *error_msg = NULL;
+		if (libssh2_session_last_error(session, &error_msg, NULL, 0) == writestate) {
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failure '%s' (%ld)", error_msg, writestate);
+		}
+
+		stream->eof = 1;
+		writestate = 0;
+	}
+
+	return writestate;
 }
 
 static size_t php_ssh2_channel_stream_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
 {
 	php_ssh2_channel_data *abstract = (php_ssh2_channel_data*)stream->abstract;
 	ssize_t readstate;
+	LIBSSH2_SESSION *session;
 
 	stream->eof = libssh2_channel_eof(abstract->channel);
 	libssh2_channel_set_blocking(abstract->channel, abstract->is_blocking);
-		
+	session = (LIBSSH2_SESSION *)zend_fetch_resource(NULL TSRMLS_CC, abstract->session_rsrc, PHP_SSH2_SESSION_RES_NAME, NULL, 1, le_ssh2_session);
+
+#ifdef PHP_SSH2_SESSION_TIMEOUT
+	if (abstract->is_blocking) {
+		libssh2_session_set_timeout(session, abstract->timeout);
+	}
+#endif
+
 	readstate = libssh2_channel_read_ex(abstract->channel, abstract->streamid, buf, count);
-	return (readstate < 0 ? 0 : readstate);
+
+#ifdef PHP_SSH2_SESSION_TIMEOUT
+	if (abstract->is_blocking) {
+		libssh2_session_set_timeout(session, 0);
+	}
+#endif
+	if (readstate == LIBSSH2_ERROR_EAGAIN) {
+		readstate = 0;
+	}
+
+	if (readstate < 0) {
+		char *error_msg = NULL;
+		if (libssh2_session_last_error(session, &error_msg, NULL, 0) == readstate) {
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failure '%s' (%ld)", error_msg, readstate);
+		}
+
+		stream->eof = 1;
+		readstate = 0;
+	}
+	return readstate;
 }
 
 static int php_ssh2_channel_stream_close(php_stream *stream, int close_handle TSRMLS_DC)
@@ -85,6 +143,18 @@ static int php_ssh2_channel_stream_set_option(php_stream *stream, int option, in
 			abstract->is_blocking = value;
 			return ret;
 			break;
+
+		case PHP_STREAM_OPTION_READ_TIMEOUT:
+			ret = abstract->timeout;
+#ifdef PHP_SSH2_SESSION_TIMEOUT
+			struct timeval tv = *(struct timeval*)ptrparam;
+			abstract->timeout = tv.tv_sec * 1000 + (tv.tv_usec / 1000);
+#else
+			php_error_docref(NULL TSRMLS_CC, E_WARNING, "No support for ssh2 stream timeout. Please recompile with libssh2 >= 1.2.9");
+#endif
+			return ret;
+			break;
+
 #if PHP_MAJOR_VERSION >= 5
 		case PHP_STREAM_OPTION_CHECK_LIVENESS:
 			return stream->eof = libssh2_channel_eof(abstract->channel);
@@ -147,7 +217,15 @@ php_url *php_ssh2_fopen_wraper_parse_path(	char *path, char *type, php_stream_co
 	if (!resource->host) {
 		return NULL;
 	}
-	
+
+	/*
+		Find resource->path in the path string, then copy the entire string from the original path.
+		This includes ?query#fragment in the path string
+	*/
+	s = resource->path;
+	resource->path = estrdup(strstr(path, resource->path));
+	efree(s);
+
 	/* Look for a resource ID to reuse a session */
 	s = resource->host;
 	if (strncmp(resource->host, "Resource id #", sizeof("Resource id #") - 1) == 0) {
@@ -444,6 +522,7 @@ static php_stream *php_ssh2_shell_open(LIBSSH2_SESSION *session, int resource_id
 	channel_data->channel = channel;
 	channel_data->streamid = 0;
 	channel_data->is_blocking = 0;
+	channel_data->timeout = 0;
 	channel_data->session_rsrc = resource_id;
 	channel_data->refcount = NULL;
 
@@ -594,15 +673,15 @@ PHP_FUNCTION(ssh2_shell)
 	int argc = ZEND_NUM_ARGS();
 
 	if (argc == 5) {
-		php_error_docref(NULL TSRMLS_CC, E_ERROR, "width specified without height paramter");
+		php_error_docref(NULL TSRMLS_CC, E_ERROR, "width specified without height parameter");
 		RETURN_FALSE;
 	}
 
 	if (zend_parse_parameters(argc TSRMLS_CC, "r|sa!lll", &zsession, &term, &term_len, &environment, &width, &height, &type) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_AUTHENTICATED_SESSION(session, zsession);
 
 	stream = php_ssh2_shell_open(session, Z_LVAL_P(zsession), term, term_len, environment, width, height, type TSRMLS_CC);
 	if (!stream) {
@@ -689,6 +768,7 @@ static php_stream *php_ssh2_exec_command(LIBSSH2_SESSION *session, int resource_
 	channel_data->channel = channel;
 	channel_data->streamid = 0;
 	channel_data->is_blocking = 0;
+	channel_data->timeout = 0;
 	channel_data->session_rsrc = resource_id;
 	channel_data->refcount = NULL;
 
@@ -816,7 +896,7 @@ PHP_FUNCTION(ssh2_exec)
 	int term_len = 0;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs|z!z!lll", &zsession, &command, &command_len, &zpty, &environment, &width, &height, &type) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	if (zpty && Z_TYPE_P(zpty) == IS_ARRAY) {
@@ -837,7 +917,7 @@ PHP_FUNCTION(ssh2_exec)
 		term_len = Z_STRLEN_P(zpty);
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_AUTHENTICATED_SESSION(session, zsession);
 
 	stream = php_ssh2_exec_command(session, Z_LVAL_P(zsession), command, term, term_len, environment, width, height, type TSRMLS_CC);
 	if (!stream) {
@@ -877,6 +957,7 @@ static php_stream *php_ssh2_scp_xfer(LIBSSH2_SESSION *session, int resource_id,
 	channel_data->channel = channel;
 	channel_data->streamid = 0;
 	channel_data->is_blocking = 0;
+	channel_data->timeout = 0;
 	channel_data->session_rsrc = resource_id;
 	channel_data->refcount = NULL;
 
@@ -950,10 +1031,10 @@ PHP_FUNCTION(ssh2_scp_recv)
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss", &zsession,  &remote_filename, &remote_filename_len, 
 																			&local_filename, &local_filename_len) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_AUTHENTICATED_SESSION(session, zsession);
 
 	remote_file = libssh2_scp_recv(session, remote_filename, &sb);
 	if (!remote_file) {
@@ -1008,10 +1089,10 @@ PHP_FUNCTION(ssh2_scp_send)
 
 	if (zend_parse_parameters(argc TSRMLS_CC, "rss|l", &zsession, &local_filename, &local_filename_len, 
 													   &remote_filename, &remote_filename_len, &create_mode) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_AUTHENTICATED_SESSION(session, zsession);
 
 	local_file = php_stream_open_wrapper(local_filename, "rb", ENFORCE_SAFE_MODE | REPORT_ERRORS, NULL);
 	if (!local_file) {
@@ -1053,15 +1134,44 @@ PHP_FUNCTION(ssh2_scp_send)
 			RETURN_FALSE;
 		}
 
-		if (bytesread != libssh2_channel_write(remote_file, buffer, bytesread)) {
-			php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failed copying file");
-			php_stream_close(local_file);
-			libssh2_channel_free(remote_file);
-			RETURN_FALSE;
+		size_t sent = 0;
+		size_t justsent = 0;
+
+		while (bytesread - sent > 0) {
+			if ((justsent = libssh2_channel_write(remote_file, (buffer + sent), bytesread - sent)) < 0) {
+
+				switch (justsent) {
+					case LIBSSH2_ERROR_EAGAIN:
+						php_error_docref(NULL TSRMLS_CC, E_WARNING, "Operation would block");
+						break;
+
+					case LIBSSH2_ERROR_ALLOC:
+						php_error_docref(NULL TSRMLS_CC,E_WARNING, "An internal memory allocation call failed");
+						break;
+
+					case LIBSSH2_ERROR_SOCKET_SEND:
+						php_error_docref(NULL TSRMLS_CC,E_WARNING, "Unable to send data on socket");
+						break;
+
+					case LIBSSH2_ERROR_CHANNEL_CLOSED:
+						php_error_docref(NULL TSRMLS_CC,E_WARNING, "The channel has been closed");
+						break;
+
+					case LIBSSH2_ERROR_CHANNEL_EOF_SENT:
+						php_error_docref(NULL TSRMLS_CC,E_WARNING, "The channel has been requested to be closed");
+						break;
+				}
+
+				php_stream_close(local_file);
+				libssh2_channel_free(remote_file);
+				RETURN_FALSE;
+			}
+			sent = sent + justsent;
 		}
 		ssb.sb.st_size -= bytesread;
 	}
 
+	libssh2_channel_flush_ex(remote_file, LIBSSH2_CHANNEL_FLUSH_ALL);
 	php_stream_close(local_file);
 	libssh2_channel_free(remote_file);
 	RETURN_TRUE;
@@ -1092,6 +1202,7 @@ static php_stream *php_ssh2_direct_tcpip(LIBSSH2_SESSION *session, int resource_
 	channel_data->channel = channel;
 	channel_data->streamid = 0;
 	channel_data->is_blocking = 0;
+	channel_data->timeout = 0;
 	channel_data->session_rsrc = resource_id;
 	channel_data->refcount = NULL;
 
@@ -1186,10 +1297,10 @@ PHP_FUNCTION(ssh2_tunnel)
 	long port;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsl", &zsession, &host, &host_len, &port) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
-	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	SSH2_FETCH_AUTHENTICATED_SESSION(session, zsession);
 
 	stream = php_ssh2_direct_tcpip(session, Z_LVAL_P(zsession), host, port TSRMLS_CC);
 	if (!stream) {
@@ -1218,7 +1329,7 @@ PHP_FUNCTION(ssh2_fetch_stream)
 	long streamid;
 
 	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rl", &zparent, &streamid) == FAILURE) {
-		RETURN_FALSE;
+		return;
 	}
 
 	if (streamid < 0) {
